# This script file provides fundamental computational functionality for 
# simulating the behaviour of the Transformer Model on a synthetic dataset.
# This script file is responsible for the initialization of the transformer 
# as well as for training the model.

# Import all required Python modules.
import torch
import torch.nn as nn
import torch.optim as optim
import os
from classes.Transformer import Transformer

# Parameters Initialization.
source_vocab_size = 100     #5000 #10
target_vocab_size = 100     #5000 #10
d_model = 200               #512  #100
num_heads = 4               #8    #4
num_layers = 4              #6    #2
d_ff = 800                  #2048 #400
max_seq_length = 10         #100  #10
batch_size = 20             #64   #20
dropout = 0.10
epochs = 500                #500

# Set the device the model will be built and executed on.
# Set the existence status of a mps GPU.
is_mps = torch.backends.mps.is_available()
# Set the existence status of a cuda GPU.
is_cuda = torch.cuda.is_available()
# Check the existence status of a mps GPU to be used during training.
if is_mps:
    device = torch.device("mps")
    print("MPS GPU is available!")
    print(70*"=")
    # Check the existence of a cuda GPU to be used during training.
elif is_cuda:
    device = torch.device("cuda")
    print("CUDA GPU is available!")
    print(70*"=")
    # Otherwise, a CPU device will be used instead.
else:
    device = torch.device("cpu")
    print("GPU is not available, CPU will be used instead!")
    print(70*"=")


# Generate random source and target data.
source_data = torch.randint(1, source_vocab_size-2, 
                            (batch_size, max_seq_length-2),device=device)
target_data = torch.randint(1, target_vocab_size-2, 
                            (batch_size, max_seq_length-2),device=device)

# In this example source and target data are considered to be word indices 
# pertaining to two different languages, namely the source language and the 
# target language. Thus, each row of the source data tensor stores a particular
# sentence in the source language while the target data tensor stores its 
# translation to the target language. However, each sentence is represented as 
# as a sequence of integers that identify the corresponding word in each 
# language. The source language vocabulary contains a source_vocab_size number 
# of different words, whereas the target language vocabulary contains a 
# target_vocab_size number of different words. Letting n=max_seq_length-2 for 
# sentences of both languages, each source and target sequence S and T stored 
# tensors source_data and target_data can be defined as follows: 
# S = {s_1,s_2,...,s_n} and T = {t_1,t_2,...,t_n}. Take into consideration that
# s_k ranges within [1..source_vocab_size-2] and 
# t_k ranges within [1..target_vocab_size-2].
# It is easy to understand that source and target vocabularies may be defined 
# in the following way:
# source_vocab = {1,...,source_vocab_size-2} U {<sos>,<eos>}
# target_vocab = {1,...,target_vocab_size-2} U {<sos>,<eos>}

# In this context, we need to denote with different symbols the starting and 
# the ending points of each sequence. Let (0) denote the start of sentence 
# symbol <sos> and (source_vocab_size-1) denote the end of sentence symbol 
# <eos>. Therefore, source and target sequences need to be enclosed in start 
# and end of sentence symbols such that: 
# S = {<sos>,s_1,s_2,...,s_n,<eos>} and T = {<sos>,t_1,t_2,...,t_n,<eos>}.

# According to the previous analysis we need to append a column with the <sos>
# to the left of the source and target data tensors. Likewise, we need to 
# append the <eos> symbol to the rigth of the source and target data tensors.

# In this example, source and target sequences are of the same length. They are
# initially drawn as sequences of integers of max_seq_length-2 from a uniform 
# probability distrivbution which generates random samples within the 
# intervals {1,...,source_vocab_size-2} and{1,...,target_vocab_size-2} 
# respectively. Evidently, enclosing each random source or target sequence 
# within the start and end of sentence symbols will ultimately form their final
# length which is equal to max_seq_length. 

# In case, however, that source and target sequences are not of the same length
# it would be critical to pad sequencs of lesser length with a designated 
# integer. In fact, this designated integer for the source sequences will be 
# identified as source_padding_idx and for the target sequences will be 
# identified as targegt_padding_idx. For this particular example we will assume
# that source_padding_idx = source_vocab_size and 
# target_padding_idx = target_vocab_size since this particular integer value
# does not appear in the original sequences.

# Set the source and target padding indices.
source_padding_idx = source_vocab_size
target_padding_idx = target_vocab_size

# Initialize the Transformer Model.
transformer = Transformer(source_vocab_size, target_vocab_size, d_model, 
                         num_heads, num_layers, d_ff, max_seq_length, 
                         source_padding_idx, target_padding_idx, dropout,
                         device)

# Move the instance of the model to the selected device.
transformer = transformer.to(device)

# Define the sos_data and end_data tensors.
sos_data = (0)*torch.ones(batch_size,1,dtype=torch.long,device=device)
eos_data = (source_vocab_size-1)*torch.ones(batch_size,1,dtype=torch.long,
                                            device=device)
# Append sos_data and eos_data tensors to the left and to the right of the 
# source_data and target_data tensors.
source_data = torch.cat((sos_data,source_data),-1)
source_data = torch.cat((source_data,eos_data),-1)
target_data = torch.cat((sos_data,target_data),-1)
target_data = torch.cat((target_data,eos_data),-1) 

# Define the fundamental training parameters of the transformer model.
# The ignore_index below specifies a target value that is ignored and does not 
# contribute to the input gradient. For this example the value to be ignored is
# the integer value which was designated to indicate the padded values within
# the target sequence.
# betas: are the coefficients used for computing running averages of gradient 
#        and its square.
# eps: term added to the denominator to improve numerical stability.
criterion = nn.CrossEntropyLoss(ignore_index=target_padding_idx)
optimizer = optim.Adam(transformer.parameters(), lr=0.0001, 
                             betas=(0.9, 0.98), eps=1e-9)

# Check the existence of a trained transformer model stored in file:
# "transformer_model.pth"
if not os.path.exists("transformer_model.pth"):
    print("Trained model file does not exist. Begining training:")
    # Train the Transformer Model.
    transformer.train()
    for epoch in range(epochs):
        optimizer.zero_grad()
        output = transformer(source_data,target_data)
        loss = criterion(output.contiguous().view(-1, target_vocab_size),
                         target_data.contiguous().view(-1))
        loss.backward()
        optimizer.step()
        # Compute the word-index based output of the transformer after training.
        final_output = torch.argmax(output,dim=2)
        # Compute the mean absolute difference between the final output and the
        # target output.
        diff = final_output - target_data
        mae = torch.sum(torch.abs(diff)) / diff.numel()
        print(f"Epoch: {epoch+1} Loss:{loss.item()} MAE:{mae.item()}")
    # Report the trainable parameters of the model.
    transformer.report_model_parameters()
    # Save the trained model.
    torch.save(transformer, 'transformer_model.pth')
else:
    print("Trained model file exists. Model will be loaded:")
    # Load the trained model.
    transformer = torch.load("transformer_model.pth")
    # Report the trainable parameters of the model.
    transformer.report_model_parameters()